T_USERNAME = r'(^[a-z0-9]{3,16}$)'
T_PASSWORD = r'(^.{5,32}$)'
T_SESSIONID = r'(^[0-9a-f]{64}$)'
T_COMMANDID = r'(^[0-9a-f]{32}$)'
T_OBJECTNAME = r'(^.{1,63}$)'
T_COMMANDID_SIZE = 32
T_WORKER_SIZE = 32
T_PARAMETERS_SIZE = 1024
T_RESULT_SIZE = 4096
T_SESSIONID_SIZE = 64
T_USERNAME_SIZE = 16
T_PID = int
